/*
 * USR2.c V0.2
 *
 * This file contains hardware driver functions for some USRobotics modems.
 * Made from compilations from the old US_Robotics driver, originally from
 * Steven wormley <wormley@step.mother.com> and the generic file IS_101.c
 * Thomas Hellstroem <thomas@vtd.volvo.se> 1996-11-14
 */

#include "../include/voice.h"

char *libvoice_US_Robotics_c = "$Id: US_Robotics.c,v 1.11 1996/11/25 19:55:16 marc Exp $";

#define BEEPLEN_DIVISOR 10

/*
 * Here we save the current mode of operation of the voice modem when
 * switching to voice mode, so that we can restore it afterwards.
 */

static char mode_save[VOICE_BUF_LEN] = "";

/*
 * Internal status variables for aborting some voice modem actions.
 */

static volatile int stop_playing;
static volatile int stop_recording;
static volatile int in_voice_mode = 0;

#define USR_BUFFER_SIZE 360
static volatile int buffer_size = USR_BUFFER_SIZE;

int USR_init (void)
     {
     char buffer[VOICE_BUF_LEN];

     watchdog_flag = FALSE;
     lprintf(L_MESG, "USR Sportster Vi 33600 voice modem");
     voice_modem_state = INITIALIZING;

     if ( voice_command("ATS30=60#VTD=3f,1,3f", "OK") != VMA_USER_1)
          lprintf(L_WARN, "voice init failed, continuing");

     sprintf(buffer, "AT#VSD=1#VSS=%d#VSP=%d",
      cvd.rec_silence_threshold.d.i * 3 / 100,
      cvd.rec_silence_len.d.i);

     if (voice_command(buffer, "OK") != VMA_USER_1)
          lprintf(L_WARN,"setting recording preferences didn't work");

     if (voice_command("AT&R2&I2", "OK") == VMA_USER_1)
          {
          TIO tio;
          tio_get(voice_fd, &tio);
          tio_set_flow_control(voice_fd, &tio, FLOW_BOTH);
          tio_set(voice_fd, &tio);
          }
     else
          lprintf(L_WARN,"can't turn on hardware flow control");

     voice_modem_state = IDLE;
     return(OK);
     }

int USR_answer_phone (void)
     {
     watchdog_flag = FALSE;

     if (voice_command("AT#VLS=4#CLS=8A", "VCON") != VMA_USER_1)
          return(VMA_ERROR);

     return(VMA_OK);
     }

int USR_beep (int frequency, int length)
     {
     char buffer[VOICE_BUF_LEN];
     sleep(1);
     watchdog_flag = FALSE;
     sprintf(buffer, "AT#VTS=[%d,0,%d]", frequency, length / BEEPLEN_DIVISOR);

     if (voice_command("AT","OK") != VMA_USER_1)
          return(FAIL);

     sleep(1);

     if (voice_command(buffer, "OK") != VMA_USER_1)
          return(FAIL);

     if (voice_command("AT","OK") != VMA_USER_1)
          return(FAIL);

     return(OK);
     }

int USR_handle_dle (char data)
     {
     watchdog_flag = FALSE;

     switch (data)
          {

          /*
           * shielded <DLE> code
           */

          case DLE:
               lprintf(L_WARN, "%s: Shielded <DLE> received", program_name);
               return(OK);

          /*
           * <ETX> code
           */

          case ETX:
               lprintf(L_WARN, "%s: <DLE> <ETX> received", program_name);
               return(OK);

          /*
           * DTMF tone detected
           */

          case '0':
          case '1':
          case '2':
          case '3':
          case '4':
          case '5':
          case '6':
          case '7':
          case '8':
          case '9':
          case '*':
          case '#':
               {
               event_type *event;
               event_data dtmf;

               event = create_event(RECEIVED_DTMF);
               dtmf.c = data;
               event->data = dtmf;
               queue_event(event);
               return(OK);
               }

          /*
           * Busy tone detected
           */

          case 'b':
               return(queue_event(create_event(BUSY_TONE)));

          /*
           * Fax calling tone detected
           */

          case 'c':
               return(queue_event(create_event(FAX_CALLING_TONE)));

          /*
           * Dial tone detected
           */

          case 'd':
               return(queue_event(create_event(DIAL_TONE)));

          /*
           * Data calling tone detected
           */

          case 'e':
               return(queue_event(create_event(DATA_CALLING_TONE)));

          case 'q':
               return(queue_event(create_event(SILENCE_DETECTED)));

          /*
           * XON received
           */

          case 'Q':
               lprintf(L_WARN, "%s: XON received", program_name);
               return(OK);

          /*
           * Ringback detected
           */

          case 'r':
               return(queue_event(create_event(RINGBACK_DETECTED)));

          /*
           * Ring detected
           */

          case 'R':
               return(queue_event(create_event(RING_DETECTED)));

          /*
           * Modem could not detect voice energy on the line
           */

          case 's':
               return(queue_event(create_event(NO_VOICE_ENERGY)));

          /*
           * XOFF received
           */

          case 'S':
               lprintf(L_WARN, "%s: XOFF received", program_name);
               return(OK);

          };

     /*
      * Unknown DLE code
      */

     lprintf(L_WARN, "%s: Unknown code <DLE> <%c>", program_name, data);
     return(FAIL);
     }

int USR_play_file (int fd)
     {
     TIO tio_save;
     TIO tio;
     char input_buffer[USR_BUFFER_SIZE];
     char output_buffer[2 * USR_BUFFER_SIZE];
     int i;
     int bytes_in;
     int bytes_out;

     watchdog_flag = FALSE;
     voice_check_events();
     stop_playing = FALSE;
     voice_modem_state = PLAYING;
     tio_get(voice_fd, &tio);
     tio_save = tio;

     if (cvd.do_hard_flow.d.i)
          tio_set_flow_control(voice_fd, &tio, FLOW_HARD);
     else
          tio_set_flow_control(voice_fd, &tio, FLOW_HARD|FLOW_XON_OUT);

     tio_set(voice_fd, &tio);

     if (voice_command("ATE0#VTX", "CONNECT") != VMA_USER_1)
          return(FAIL);

     while (!stop_playing)
          {
          watchdog_flag = FALSE;

          if ((bytes_in = read(fd, input_buffer, buffer_size)) <= 0)
               break;

          bytes_out = 0;

          for(i = 0; i < bytes_in; i++)
               {
               output_buffer[bytes_out] = input_buffer[i];

               if (output_buffer[bytes_out++] == DLE)
                    output_buffer[bytes_out++] = DLE;

               };

          lprintf(L_JUNK, "%s: <DATA %d bytes>", program_name, bytes_out);

          if (voice_write_raw(output_buffer, bytes_out) != OK)
               return(FAIL);

          while (check_for_input(voice_fd))
               {
               char modem_byte;

               if ((modem_byte = voice_read_char()) == FAIL)
                    return(FAIL);

               if (modem_byte == DLE)
                    {

                    if ((modem_byte = voice_read_char()) == FAIL)
                         return(FAIL);

                    lprintf(L_JUNK, "%s: <DLE> <%c>", voice_modem_name,
                     modem_byte);
                    voice_modem->handle_dle(modem_byte);
                    }
               else
                    lprintf(L_WARN, "%s: unexpected byte %c from voice modem",
                     program_name, modem_byte);

               };

          voice_check_events();
          };

     if (stop_playing)
          {
          sprintf(output_buffer, "%c%c", DLE, CAN);
          lprintf(L_JUNK, "%s: <DLE> <CAN>", program_name);
          }
     else
          {
          sprintf(output_buffer, "%c%c", DLE, ETX);
          lprintf(L_JUNK, "%s: <DLE> <ETX>", program_name);
          };

     if (voice_write_raw(output_buffer, strlen(output_buffer)) != OK)
          return(FAIL);

     if ((voice_command("", "OK|VCON") & VMA_USER) != VMA_USER)
          return(FAIL);

     if (voice_write_raw("ATE1\r",5) != OK)
          return(FAIL);

     if ((voice_command("", "OK|VCON") & VMA_USER) != VMA_USER)
          return(FAIL);

     tio_set(voice_fd, &tio_save);

     if (voice_command("AT", "OK") != VMA_USER_1)
          return(FAIL);

     voice_modem_state = IDLE;
     voice_check_events();

     if (stop_playing)
          return(INTERRUPTED);

     return(OK);
     }

int USR_record_file (int fd)
     {
     TIO tio_save;
     TIO tio;
     time_t timeout;
     char input_buffer[USR_BUFFER_SIZE];
     char output_buffer[USR_BUFFER_SIZE];
     int i = 0;
     int bytes_in = 0;
     int bytes_out;
     int got_DLE_ETX = FALSE;
     int was_DLE = FALSE;

     watchdog_flag = FALSE;
     voice_check_events();
     timeout = time(NULL) + cvd.rec_max_len.d.i;
     stop_recording = FALSE;
     voice_modem_state = RECORDING;
     tio_get(voice_fd, &tio);
     tio_save = tio;

     if (cvd.do_hard_flow.d.i)
          tio_set_flow_control(voice_fd, &tio, FLOW_HARD);
     else
          tio_set_flow_control(voice_fd, &tio, FLOW_HARD | FLOW_XON_IN);

     tio.c_cc[VMIN] = (buffer_size > 0xff) ? 0xff : buffer_size;
     tio.c_cc[VTIME] = 1;
     tio_set(voice_fd, &tio);

     if (voice_command("AT#VRX", "CONNECT") != VMA_USER_1)
          return(FAIL);

     while (!got_DLE_ETX)
          {
          watchdog_flag = FALSE;

          if (timeout < time(NULL))
               voice_stop_recording();

          if ((bytes_in = read(voice_fd, input_buffer, buffer_size)) <= 0)
               {
               lprintf(L_ERROR, "%s: could not read byte from voice modem",
                program_name);
               return(FAIL);
               };

          bytes_out = 0;

          for (i = 0; (i < bytes_in) && !got_DLE_ETX; i++)
               {

               if (was_DLE)
                    {
                    was_DLE = FALSE;

                    switch (input_buffer[i])
                         {
                         case DLE:
                              output_buffer[bytes_out++] = DLE;
                              break;
                         case ETX:
                              got_DLE_ETX = TRUE;
                              lprintf(L_JUNK, "%s: <DATA %d bytes>",
                               voice_modem_name, bytes_out);
                              lprintf(L_JUNK, "%s: <DLE> <ETX>",
                               voice_modem_name);
                              break;
                         default:
                              lprintf(L_JUNK, "%s: <DLE> <%c>",
                               voice_modem_name, input_buffer[i]);
                              voice_modem->handle_dle(input_buffer[i]);
                         };

                    }
               else
                    {

                    if (input_buffer[i] == DLE)
                         was_DLE = TRUE;
                    else
                         output_buffer[bytes_out++] = input_buffer[i];

                    };

               };

          write(fd, output_buffer, bytes_out);

          if (!got_DLE_ETX)
               lprintf(L_JUNK, "%s: <DATA %d bytes>", voice_modem_name,
                bytes_out);

          voice_check_events();
          };

     tio_set(voice_fd, &tio_save);

     if ((voice_analyze(&input_buffer[i], "\r\nOK|\r\nVCON") & VMA_USER) ==
      VMA_USER)
          lprintf(L_JUNK, "%s: OK|VCON", voice_modem_name);
     else
          {

          /*
           * Fixme: The datas in the buffer should be checked
           * for DLE shielded codes
           */

          int j;

          lprintf(L_JUNK, "%s: data left in buffer: ",
           voice_modem_name);

          for (j = i; j < bytes_in ; j++)
               lputc(L_JUNK, input_buffer[j]);

          };

     if (voice_command("AT", "OK") != VMA_USER_1)
          return(FAIL);

     voice_modem_state = IDLE;
     voice_check_events();
     return(OK);
     }

int USR_set_buffer_size (int size)
     {

     if (size > USR_BUFFER_SIZE)
          return(FAIL);

     buffer_size = size;
     return(OK);
     }

int USR_set_compression(int *compression, int *speed, int *bits)
     {

     if (*compression == 0)
          *compression = 2;

     if (*speed == 0)
          *speed = 7200;

     if (*speed != 7200)
          {
          lprintf(L_WARN, "%s: Illegal sample rate (%d)",
           voice_modem_name, *speed);
          return(FAIL);
          };

     switch (*compression)
          {
          case 2:
               *bits = 2;
               buffer_size = USR_BUFFER_SIZE * 2 / 4;

               if (voice_command("AT#VBS=2", "OK") != VMA_FAIL)
                    return(OK);
               else
                    return(VMA_FAIL);

          case 3:
               *bits = 3;
               buffer_size = USR_BUFFER_SIZE * 3 / 4;

               if (voice_command("AT#VBS=3", "OK") != VMA_FAIL)
                    return(OK);
               else
                    return(VMA_FAIL);

          case 4:
               *bits = 4;
               buffer_size = USR_BUFFER_SIZE * 4 / 4;

               if (voice_command("AT#VBS=4", "OK") != VMA_FAIL)
                    return(OK);
               else
                    return(VMA_FAIL);

          };

     lprintf(L_WARN,
      "USR SPVi handle event: Illeagal voice compression method (%d)",
      *compression);
     return(FAIL);
     }

int USR_set_device(int device)
     {
     watchdog_flag = FALSE;

     switch (device)
          {
          case NO_DEVICE:
               voice_command("AT#VLS=0","OK|VCON");
               return(OK);
          case DIALUP_LINE:
               voice_command("AT#VLS=0","OK|VCON");
               return(OK);
          case EXTERNAL_MICROPHONE:
               voice_command("AT#VLS=1","OK|VCON");
               return(OK);
          case INTERNAL_SPEAKER:
               voice_command("AT#VLS=2","OK|VCON");
               return(OK);
          case INTERNAL_MICROPHONE:
               voice_command("AT#VLS=3","OK|VCON");
               return(OK);
          default:
               lprintf(L_WARN,"USR SPVi handle event: Unknown output device (%d)",
                device);
               return(FAIL);
          };

     }

int USR_stop_recording (void)
     {
     watchdog_flag = FALSE;
     stop_recording = TRUE;

     if (voice_write_raw("AT", 2) != OK)
          return(FAIL);

     return(OK);
     }

int USR_voice_mode_off (void)
     {
     char buffer[VOICE_BUF_LEN];

     watchdog_flag = FALSE;
     sprintf(buffer, "AT#CLS=%s", mode_save);

     if (voice_command(buffer, "OK") != VMA_USER_1)
          return(FAIL);

     in_voice_mode = 0;
     return(OK);
     }

int USR_voice_mode_on (void)
     {
     watchdog_flag = FALSE;

     if (voice_command("AT#CLS?", "") != OK)
          return(FAIL);

     if (voice_read(mode_save) != OK)
          return(FAIL);

     if (voice_command("", "OK") != VMA_USER_1)
          return(FAIL);

     if (voice_command("AT#CLS=8", "OK") != VMA_USER_1)
          return(FAIL);

     in_voice_mode = -1;
     return(OK);
     }

int USR_stop_playing (void)
     {
     watchdog_flag = FALSE;
     stop_playing = TRUE;
     return(OK);
     }

int USR_switch_to_data_fax (char *mode)
     {
     char buffer[VOICE_BUF_LEN];

     watchdog_flag = FALSE;

     if (in_voice_mode)
          {

          if(voice_mode_off() == FAIL)
                return(FAIL);

          };

     sprintf(buffer, "AT+FCLASS=%s", mode);

     if (voice_command(buffer, "OK") != VMA_USER_1)
          return(FAIL);

     return(OK);
     }


voice_modem_struct US_Robotics =
     {
     "US Robotics",
     "US Robotics",
     &USR_answer_phone,
     &USR_beep,
     &IS_101_dial,
     &USR_handle_dle,
     &USR_init,
     &IS_101_message_light_off,
     &IS_101_message_light_on,
     &USR_play_file,
     &USR_record_file,
     &USR_set_compression,
     &USR_set_device,
     &IS_101_stop_dialing,
     &USR_stop_playing,
     &USR_stop_recording,
     &IS_101_stop_waiting,
     &USR_switch_to_data_fax,
     &USR_voice_mode_off,
     &USR_voice_mode_on,
     &IS_101_wait
     };
